Attribute VB_Name = "modCryptDecrypt"
' // modCryptDecrypt.bas - module for encryption and decryption module
' //  Krivous Anatoly Anatolevich (The trick), 2016

Option Explicit

Private Type SAFEARRAYBOUND
    cElements               As Long
    lLBound                 As Long
End Type

Private Type SAFEARRAY
    cDims                   As Integer
    fFeatures               As Integer
    cbElements              As Long
    cLocks                  As Long
    pvData                  As Long
    rgsabound               As SAFEARRAYBOUND
End Type

Private Type IMAGE_SECTION_HEADER
    SectionName(7)          As Byte
    VirtualSize             As Long
    VirtualAddress          As Long
    SizeOfRawData           As Long
    PointerToRawData        As Long
    PointerToRelocations    As Long
    PointerToLinenumbers    As Long
    NumberOfRelocations     As Integer
    NumberOfLinenumbers     As Integer
    Characteristics         As Long
End Type

Private Declare Function ArrPtr Lib "msvbvm60" _
                         Alias "VarPtr" ( _
                         ByRef arr() As Any) As Long
Private Declare Function PutMem4 Lib "msvbvm60" ( _
                         ByVal pDst As Long, _
                         ByVal NewValue As Long) As Long
Private Declare Function CopyMemory Lib "kernel32" _
                         Alias "RtlMoveMemory" ( _
                         ByRef Destination As Any, _
                         ByRef Source As Any, _
                         ByVal Length As Long) As Long
Private Declare Function GetModuleFileName Lib "kernel32" _
                         Alias "GetModuleFileNameW" ( _
                         ByVal hModule As Long, _
                         ByVal lpFileName As Long, _
                         ByVal nSize As Long) As Long
Private Declare Function CommandLineToArgvW Lib "Shell32" ( _
                         ByVal lpCmdLine As Long, _
                         ByRef pNumArgs As Long) As Long
Private Declare Function lstrlen Lib "kernel32" _
                         Alias "lstrlenW" ( _
                         ByVal lpString As Long) As Long
Private Declare Function lstrcpyn Lib "kernel32" _
                         Alias "lstrcpynW" ( _
                         ByVal lpString1 As Long, _
                         ByVal lpString2 As Long, _
                         ByVal iMaxLength As Long) As Long
Private Declare Function GlobalFree Lib "kernel32" ( _
                         ByVal hMem As Long) As Long
Private Declare Function GetMem4 Lib "msvbvm60" ( _
                         ByRef Src As Any, _
                         ByRef Dst As Any) As Long
Private Declare Function GetMem2 Lib "msvbvm60" ( _
                         ByRef Src As Any, _
                         ByRef Dst As Any) As Long
Private Declare Function VirtualProtect Lib "kernel32" ( _
                         ByVal lpAddress As Long, _
                         ByVal dwSize As Long, _
                         ByVal flNewProtect As Long, _
                         ByRef lpflOldProtect As Long) As Long
                         
Private Const PAGE_EXECUTE_READWRITE  As Long = &H40

Private mapDesc As SAFEARRAY

' ///////////////////////////////////////////////////////////////////////////////////
' ////////////////////////// THIS REGION GET NEVER ENCRYPT //////////////////////////
' ///////////////////////////////////////////////////////////////////////////////////

' // Code placed between BEGIN_OF_NON_ENCRYPTABLE_REGION and
' // END_OF_NON_ENCRYPTABLE_REGION functions is not encrypt

Private Function BEGIN_OF_NON_ENCRYPTABLE_REGION() As Long
    BEGIN_OF_NON_ENCRYPTABLE_REGION = 1
End Function

' // Parse command line
Private Function ParseCommandLine( _
                 ByRef Line As String, _
                 ByRef Out() As String) As Boolean
    Dim ptr     As Long
    Dim count   As Long
    Dim index   As Long
    Dim strLen  As Long
    Dim strAdr  As Long
    
    ptr = CommandLineToArgvW(StrPtr(Line), count)
    
    If count < 1 Then Exit Function
    
    ReDim Out(count - 1)
    
    For index = 0 To count - 1
        GetMem4 ByVal ptr + index * 4, strAdr
        strLen = lstrlen(strAdr)
        Out(index) = Space(strLen)
        lstrcpyn StrPtr(Out(index)), strAdr, strLen + 1
    Next
    
    GlobalFree ptr
    
    ParseCommandLine = True
    
End Function

' // Map an array to data
Private Sub MapArray( _
            ByRef arr() As Byte, _
            ByVal lpPointer As Long, _
            ByVal size As Long)
            
    mapDesc.cDims = 1
    mapDesc.cbElements = 1
    mapDesc.rgsabound.lLBound = 0
    mapDesc.rgsabound.cElements = size
    mapDesc.pvData = lpPointer
    
    PutMem4 ArrPtr(arr()), VarPtr(mapDesc)
    
End Sub

' // Unmap array
Private Sub UnmapArray( _
            ByRef arr() As Byte)
    PutMem4 ArrPtr(arr()), 0&
End Sub

' // Determines whether exe is encrypted or not
Private Function IsEncrypted() As Boolean
    Dim ptr         As Long
    Dim hModule     As Long
    Dim Checksum    As Long
    
    ' // Get VBHeader.pProjectInfo
    hModule = App.hInstance
    ' // Get e_lfanew
    GetMem4 ByVal hModule + &H3C, ptr
    ' // Get AddressOfEntryPoint
    GetMem4 ByVal ptr + &H28 + hModule, ptr
    ' // Get VBHeader
    GetMem4 ByVal ptr + hModule + 1, ptr
    ' // Get VBHeader.pProjectInfo
    GetMem4 ByVal ptr + &H30, ptr
    ' // Get checksum (VbProjectInfo.dwNull)
    GetMem4 ByVal ptr + 8, Checksum
    
    IsEncrypted = Checksum < 0
    
End Function

' // Get information about code location
Private Function GetPlacementData( _
                 ByRef hModule As Long, _
                 ByRef pStartCode As Long, _
                 ByRef pEndCode As Long, _
                 ByRef uStart As Long, _
                 ByRef uEnd As Long, _
                 ByRef sections() As IMAGE_SECTION_HEADER, _
                 ByRef pChecksum As Long) As Boolean
    Dim ptr         As Long
    Dim numOfSect   As Integer
    Dim dwSize      As Long
    
    ' // Get VBHeader.pProjectInfo
    hModule = App.hInstance
    ' // Get e_lfanew
    GetMem4 ByVal hModule + &H3C, ptr
    ' // Get AddressOfEntryPoint
    GetMem4 ByVal ptr + &H28 + hModule, ptr
    ' // Get VBHeader
    GetMem4 ByVal ptr + hModule + 1, ptr
    ' // Get VBHeader.pProjectInfo
    GetMem4 ByVal ptr + &H30, ptr
    ' // Get checksum (VbProjectInfo.dwNull)
    pChecksum = ptr + 8
    ' // Get Start of code
    GetMem4 ByVal ptr + &HC, pStartCode
    ' // Get End of code
    GetMem4 ByVal ptr + &H10, pEndCode
    
    ' // Get sections
    ' // Get e_lfanew
    GetMem4 ByVal hModule + &H3C, ptr
    ' // Get IMAGE_FILE_HEADER.NumberOfSections
    GetMem2 ByVal ptr + &H6 + hModule, numOfSect
    ' // Get sections
    ReDim sections(numOfSect - 1)
    ' // Get size of IMAGE_FILE_HEADER.SizeOfOptionalHeader
    GetMem2 ByVal ptr + &H14 + hModule, dwSize
    
    ptr = hModule + ptr + &H18 + dwSize
    CopyMemory sections(0), ByVal ptr, Len(sections(0)) * numOfSect
    
    ' // Get non-encryption region
    uStart = GetAddr(AddressOf BEGIN_OF_NON_ENCRYPTABLE_REGION)
    uEnd = GetAddr(AddressOf END_OF_NON_ENCRYPTABLE_REGION)
    
    GetPlacementData = True
    
End Function

' // Decrypt exe
Private Function DecryptEXE( _
                 ByRef bytLicense() As Byte) As Boolean
    Dim bytHash()       As Byte:                    Dim hIndex          As Long
    Dim lIndex          As Long:                    Dim cIndex          As Long
    Dim hModule         As Long:                    Dim pStartCode      As Long
    Dim pEndCode        As Long:                    Dim uStart          As Long
    Dim uEnd            As Long:                    Dim numOfSect       As Integer
    Dim sections()      As IMAGE_SECTION_HEADER:    Dim dwSize          As Long
    Dim code()          As Byte:                    Dim Checksum        As Long
    Dim pChecksum       As Long:                    Dim origSum         As Long
    Dim bRet            As Boolean
    
    ' // Get placement data
    GetPlacementData hModule, pStartCode, pEndCode, uStart, uEnd, sections(), pChecksum
    
    dwSize = pEndCode - pStartCode + 1
    
    ' // Allow to write
    VirtualProtect pStartCode, dwSize, PAGE_EXECUTE_READWRITE, 0
    ' // Map array to exe
    MapArray code(), pStartCode, dwSize
    
    ' // Get original checksum
    GetMem4 ByVal pChecksum, origSum
    
    ' // If last bit is set exe is encrypted
    If origSum < 0 Then
        
        ' // Reset last bit
        origSum = origSum And &H7FFFFFFF
        
        lIndex = 0: cIndex = 0
        ' // Get indices of uncryptable region
        uStart = uStart - pStartCode: uEnd = uEnd - pStartCode
        
        ' // Decrypt
        Do While cIndex < dwSize
            
            If cIndex = uStart Then
                cIndex = cIndex + (uEnd - uStart)
            Else
                
                code(cIndex) = code(cIndex) Xor bytLicense(lIndex)
                Checksum = ((Checksum * 123) Xor code(cIndex)) And &HFFFFFFF
                
                lIndex = lIndex + 1
                
                If lIndex > UBound(bytLicense) Then lIndex = 0
                
            End If
    
            cIndex = cIndex + 1
            
        Loop
        
        ' // Check checksum
        If origSum = Checksum Then bRet = True
    
    Else: bRet = True
    
    End If
    
    UnmapArray code()
    
    DecryptEXE = bRet
    
End Function

Public Sub Main()
    Dim blInIDE     As Boolean
    Dim cmdLine     As String
    Dim args()      As String
    Dim fileName    As String
    Dim fNum        As Integer
    Dim license()   As Byte
    
    ' // Determine whether the application is compiled or not
    Debug.Assert MakeTrue(blInIDE)
    
    ' // Works only in compiled form
    If Not blInIDE Then
        
        cmdLine = Command$()

        ' // Check command line
        If ParseCommandLine(cmdLine, args()) Then
            If UBound(args) = 0 Then
                ' // Check param. It should be "crypt:filename"
                ' // You can call it only one time, because EncryptMe function also encrypted
                If StrComp(Left$(args(0), 6), "crypt:", vbTextCompare) = 0 Then
                    
                    ' // Check if encrypted
                    If IsEncrypted() Then
                        MsgBox "Module has already been encrypted", vbExclamation
                        Exit Sub
                    End If
                    
                    ' // Crypt exe
                    fileName = Mid$(args(0), 7)
                    
                    If EncryptMe(fileName) Then
                        MsgBox "Encrypt successfull!"
                    Else
                        MsgBox "Encrypt unsuccessfull", vbCritical
                    End If
                    
                    Exit Sub
                    
                End If
            End If
            
        End If
        
        ' // Try to decrypt EXE
        ' // 1. Check license file
        fileName = App.Path & "\license.lic"
        
        If Len(Dir$(fileName)) = 0 Then
            ' // License not found. Show window with key info
            MsgBox "License file not found", vbExclamation
            ' // Exit
            Exit Sub
        Else
            
            ' // Open license
            fNum = FreeFile
            
            Open fileName For Binary As fNum
            ReDim license(LOF(fNum) - 1)
            Get fNum, , license()
            Close fNum
            
            ' // Try to decrypt
            If Not DecryptEXE(license()) Then
                ' // Unsuccessful
                MsgBox "Invalid license", vbExclamation
                ' // Exit
                Exit Sub
            End If
            
        End If
        
    End If
    
    frmMain.Show
    
End Sub

Private Function GetAddr( _
                 ByVal addr As Long) As Long
    GetAddr = addr
End Function

Private Function MakeTrue( _
                 ByRef blTrue As Boolean) As Boolean
    blTrue = True
    MakeTrue = True
End Function

Private Function RVA2RAW( _
                 ByRef sections() As IMAGE_SECTION_HEADER, _
                 ByVal rva As Long) As Long
    Dim i As Long

    For i = 0 To UBound(sections)
        If rva >= sections(i).VirtualAddress And _
           rva < sections(i).VirtualAddress + sections(i).VirtualSize Then
            RVA2RAW = sections(i).PointerToRawData + (rva - sections(i).VirtualAddress)
            Exit Function
        End If
    Next
    
    RVA2RAW = rva
    
End Function

Private Function END_OF_NON_ENCRYPTABLE_REGION() As Long
    END_OF_NON_ENCRYPTABLE_REGION = 2
End Function

' ///////////////////////////////////////////////////////////////////////////////////
' ///////////////////////////// THIS REGION GET ENCRYPT /////////////////////////////
' ///////////////////////////////////////////////////////////////////////////////////

' // Encrypt exe and store it as [ModuleName]__encrypted.[extension] into same path
Private Function EncryptMe( _
                 ByRef fileName As String) As Boolean
    Dim hModule         As Long:                    Dim pStartCode      As Long
    Dim pEndCode        As Long:                    Dim crDat()         As Byte
    Dim exeDat()        As Byte:                    Dim fNum            As Integer
    Dim uStart          As Long:                    Dim uEnd            As Long
    Dim index           As Long:                    Dim numOfSect       As Integer
    Dim sections()      As IMAGE_SECTION_HEADER:    Dim dwSize          As Long
    Dim fileOfst        As Long:                    Dim pszEXE          As String
    Dim extension       As String:                  Dim Checksum        As Long
    Dim pChecksum       As Long
    
    ' // Get placement
    GetPlacementData hModule, pStartCode, pEndCode, uStart, uEnd, sections(), pChecksum
    
    ' // Open key file
    fNum = FreeFile
    
    Open fileName For Binary As fNum
    ReDim crDat(LOF(fNum) - 1)
    Get fNum, , crDat()
    Close fNum
    
    ' // Get exe path
    pszEXE = Space$(260)
    dwSize = GetModuleFileName(hModule, StrPtr(pszEXE), Len(pszEXE))
    pszEXE = Left$(pszEXE, dwSize)
    
    ' // Open exe
    Open pszEXE For Binary As fNum
    ReDim exeDat(LOF(fNum) - 1)
    Get fNum, , exeDat()
    Close fNum
    
    ' // Start encryption
    Do While pStartCode <= pEndCode
        
        If pStartCode = uStart Then
            pStartCode = pStartCode + (uEnd - uStart)
        Else
            
            fileOfst = RVA2RAW(sections(), pStartCode - hModule)
            
            Checksum = ((Checksum * 123) Xor exeDat(fileOfst)) And &HFFFFFFF
            exeDat(fileOfst) = exeDat(fileOfst) Xor crDat(index)
            
            index = index + 1
            
            If index > UBound(crDat) Then index = 0
            
        End If
        
        pStartCode = pStartCode + 1

    Loop
    
    ' // Store checksum
    Checksum = Checksum Or &H80000000
    fileOfst = RVA2RAW(sections(), pChecksum - hModule)
    GetMem4 Checksum, exeDat(fileOfst)
    
    ' // Get extension
    dwSize = InStrRev(pszEXE, ".")
    
    If dwSize Then
    
        extension = Mid$(pszEXE, dwSize)
        pszEXE = Left$(pszEXE, dwSize - 1)
        
    End If
    
    ' // Save encrypted exe
    Open pszEXE & "_encrypted" & extension For Binary As fNum
    Put fNum, , exeDat
    Close fNum
    
    EncryptMe = True
    
End Function

